# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import yaml
import argparse
from typing import Any

from os.path import join
from os import makedirs
import subprocess


def get_git_revision_hash() -> str:
    return subprocess.check_output(
        ['git', 'rev-parse', 'HEAD']
    ).decode('ascii').strip()


def convert_float(item: Any) -> Any:
    try:
        return float(item)
    except ValueError:
        return item


def convert_str_to_float(dictionary: dict) -> None:
    for key, value in dictionary.items():
        if isinstance(value, str):
            dictionary[key] = convert_float(value)
        elif isinstance(value, list):
            dictionary[key] = [convert_float(item) if isinstance(item, str)
                               else item for item in value]
        elif isinstance(value, dict):
            convert_str_to_float(value)


def load_and_save_config(default: str) -> dict:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "config",
        help="Config File",
        type=str,
        nargs='?',
        const=1,
        default=default)
    parser.add_argument(
        '--cpu',
        action='store_true',
        help='Use CPU',
        default=False
    )
    args = parser.parse_args()

    if args.cpu:
        import jax.config
        jax.config.update('jax_platform_name', 'cpu')

    cfg_file = args.config
    with open(cfg_file, 'r') as file:
        cfg = yaml.safe_load(file)

    convert_str_to_float(cfg)
    cfg['git_hash'] = get_git_revision_hash()

    makedirs(cfg['outputdir'], exist_ok=True)
    with open(join(cfg['outputdir'], 'config.yaml'), 'w') as file:
        yaml.dump(cfg, file)

    return cfg



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import numpy as np


def calculate_critical_caustic(lens_system, mean, plot=False):
    # From Meneghetti, 2021: Introduction to Gravitational Lensing, see p.80
    '''Returns the coordinates of the critical caustic of the lens system in
    arcsec.

    Returns:
        critial: Tuple[ArrayLike, ArrayLike] = radial, tangential
        caustic: Tuple[ArrayLike, ArrayLike] = radial, tangential

    For plotting:
        lens_image.plot(*critial, 'k')
        source_image.plot(*caustic, 'k')
    '''

    import matplotlib.pyplot as plt
    from scipy.ndimage import map_coordinates
    from charm_lensing.linear_operators import DeflectionAngle

    DA = DeflectionAngle(lens_system.lens_plane_model.space.extend())
    kappa = lens_system.lens_plane_model.convergence_model.model(mean)

    arcsec = lens_system.lens_plane_model.space.distances[0]
    space = lens_system.lens_plane_model.space.extend()
    a1, a2 = DA(kappa).real
    psi12, psi11 = np.gradient(a1, arcsec)
    psi22, psi21 = np.gradient(a2, arcsec)

    gamma1 = 0.5 * (psi11 - psi22)
    gamma2 = psi12
    gamma = np.sqrt(gamma1**2+gamma2**2)
    lambdat = 1.0-kappa-gamma
    lambdar = 1.0-kappa+gamma
    detA = lambdat*lambdar

    cs = plt.contour(detA, levels=[0.0])
    plt.close()

    fig, ax = plt.subplots(1, 2, figsize=(18, 8))
    contour = cs.collections[0]
    # p contains the paths of each individual critical line
    p = contour.get_paths()
    sizevs = np.empty(len(p), dtype=int)

    # if we find any critical line, then we process it if (sizevs.size >0):
    critical, caustic = [], []
    for j in range(len(p)):
        # the contours are ensembles of polygons, called paths
        # for each path, we create two vectors containing the x1 and x2
        # coordinates of the vertices
        vs = contour.get_paths()[j].vertices
        sizevs[j] = len(vs)
        x1 = []
        x2 = []
        for i in range(len(vs)):
            xx1, xx2 = vs[i]
            x1.append(float(xx1))
            x2.append(float(xx2))
        x1 = np.array(x1)
        x2 = np.array(x2)

        a_1 = map_coordinates(a1, [[x2], [x1]], order=1)[0]
        a_2 = map_coordinates(a2, [[x2], [x1]], order=1)[0]

        # now we use the lens equation to # obtain the caustics and transform
        # the coordinates to physical units:
        offset = (-space.extent[1]+space.distances[0]/2,
                  -space.extent[3]+space.distances[1]/2)
        x1, x2 = (
            np.array(x1)*space.distances[0]+offset[0],
            np.array(x2)*space.distances[1]+offset[1]
        )
        y1 = x1-a_1
        y2 = x2-a_2
        ax[0].plot(x1, x2, '-')
        ax[1].plot(y1, y2, '-')

        critical.append(np.array((x1, x2)))
        caustic.append(np.array((y1, y2)))

    if plot:
        plt.show()
    else:
        plt.close()

    return critical, caustic



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import numpy as np
from scipy.interpolate import RectBivariateSpline


def source_distortion_ratio(input_source, model_source):
    return 10 * np.log10(np.linalg.norm(input_source) /
                         np.linalg.norm(input_source - model_source))


def chi2(data, model, std):
    return np.nanmean(((data - model)/std)**2)


def percent_inside(x, value):
    return np.sum(np.abs(value) < x)/len(value.flatten())


def shift_and_resize(input, recon, full_info=False, sxsy=None):
    ''' Shifts and resizes the reconstructed image to match the input image.
    This is used for plotting mock data.
    '''
    if input.shape != recon.shape:

        # from charm_lensing.interpolation import build_interpolation, build_xy_jnp

        # from charm_lensing.spaces import get_xycoords
        # from functools import partial

        # length = recon.shape[0]
        # recon_dist = np.array([length/s for s in recon.shape])
        # input_dist = np.array([length/s for s in input.shape])
        # input_xycoords = get_xycoords(input.shape, input_dist)
        # recon_xycoords = get_xycoords(recon.shape, recon_dist)

        # bxy = partial(build_xy_jnp, recon.shape, recon_dist)
        # inter = build_interpolation(bxy, recon_dist)
        # recon = inter(
        #     input_xycoords.reshape(2, -1) - shift,
        #     recon.reshape(-1)
        # ).reshape(input.shape)
        # return recon

        xycoord = np.linspace(0, 1, num=recon.shape[0])
        Recaster = RectBivariateSpline(
            xycoord, xycoord, recon
        )
        xycoordnew = np.linspace(0, 1, num=input.shape[0])
        recon = Recaster(*(xycoordnew,) * 2, grid=True) * (
            recon.shape[0] / input.shape[0]) ** 2

    if sxsy is None:
        # Compute the cross-correlation between the real and reconstruction arrays
        cross_correlation = np.fft.ifft2(
            np.fft.fft2(input).conj() * np.fft.fft2(recon)
        ).real

        # Find the indices where the cross-correlation is maximum
        shift_y, shift_x = np.unravel_index(
            np.argmax(cross_correlation), cross_correlation.shape)

        # Adjust the shifts if they are larger than half of the dimensions
        if shift_x > input.shape[1] // 2:
            shift_x -= input.shape[1]
        if shift_y > input.shape[0] // 2:
            shift_y -= input.shape[0]

        sx, sy = shift_y, shift_x

    else:
        sx, sy = sxsy

    recon = np.roll(recon, (-sx, -sy), axis=(0, 1))

    if full_info:
        return recon, sx, sy
    return recon



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import jax.numpy as jnp


def rotation(grid, theta):
    """
    Rotates the passed coordinates anticlockwise by an angle theta
    """
    x = grid[0] * jnp.cos(theta) - grid[1] * jnp.sin(theta)
    y = grid[0] * jnp.sin(theta) + grid[1] * jnp.cos(theta)
    return x, y



from charm_lensing.utils import from_random, full

from charm_lensing.response import (
    build_integration_operator, build_zero_flux)
from charm_lensing.spaces import Space

import numpy as np


def test_integration_operator():
    domain = Space((256, 256), (1, 1))
    target_distances = (2, 2)

    integration_operator = build_integration_operator(
        domain,
        target_distances,
        oversampling=2
    )
    from jax import jit
    j_integration_operator = jit(integration_operator)

    f = np.full(domain.shape, 1.)
    h = j_integration_operator(f)

    assert np.allclose(np.full((128, 128), 4), h)


def test_zero_flux():
    p = {
        'zero_flux': {
            'mean': {
                'distribution': 'normal',
                'mean': 0.1,
                'sigma': 0.01}
        },
        'randim': {
            'distribution': 'uniform',
            'min': 0,
            'max': 1
        }
    }

    ZeroFlux = build_zero_flux('l1', p)
    pos = full(ZeroFlux.domain, 0.0)
    out = ZeroFlux(pos)

    assert np.allclose(out, 0.1)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .rotation import rotation

import jax.numpy as jnp
from jax import custom_jvp

from typing import Tuple
from numpy.typing import ArrayLike


@custom_jvp
def F(x):
    def bigger(x):
        return (x**2-1)**(-1/2) * jnp.arctan(jnp.sqrt(x**2-1))

    def smaller(x):
        return (1-x**2)**(-1/2) * jnp.arctanh(jnp.sqrt(1-x**2))

    x0 = jnp.where(x > 1, bigger(x), 0)
    x1 = jnp.where(x < 1, smaller(x), 0)
    x2 = jnp.where(x == 1, 1, 0)

    return x0 + x1 + x2


@F.defjvp
def F_jvp(primals, tangents):
    x, = primals
    x_dot, = tangents
    primal_out = F(x)
    tangent_out = jnp.where(x != 1, (1 - x**2 * F(x)) / (x*(x**2 - 1)), 0)
    return primal_out, tangent_out * x_dot


class Nfw:
    prior_keys = (('b', (1,)), ('rs', (1,)), ('center', (2,)),
                  ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        # convergence field (Keeton 2002 eq. 55)
        b, rs, center, theta, q = params

        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)
        R = jnp.hypot(q*x, y)

        x = R / rs
        return 2*b*(1-F(x))/(x**2-1)


class Cnfw:
    prior_keys = (('b', (1,)), ('rs', (1,)), ('center', (2,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        # convergence field (Keeton 2002 eq. 55)
        b, rs, center = params

        x, y = coords[0]-center[0], coords[1]-center[1]
        R = jnp.hypot(x, y)

        x = R / rs
        return 2*b*(1-F(x))/(x**2-1)

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        # deflection field (Keeton 2002 eq. 56)
        b, rs, center = params

        x, y = coords[0]-center[0], coords[1]-center[1]
        R = jnp.hypot(x, y)
        theta = jnp.arctan2(y, x)

        x = R / rs
        alpha = 4*b*rs*(jnp.log(x/2)+F(x))/x

        sol = alpha * jnp.exp(1j * theta)
        return jnp.array((sol.real, sol.imag))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import numpy as np
from numpy.typing import ArrayLike
from typing import Any, Optional, Callable

from scipy.interpolate import RectBivariateSpline
from scipy.optimize import brentq

import nifty8 as ift

from charm_lensing.spaces import get_centered_slice
from charm_lensing.spaces import coords
from charm_lensing.utils import load_fits
from charm_lensing.space_model import SpaceModel
from charm_lensing.interpolation import build_linear_interpolation
from charm_lensing.utils import estimate_snr
from charm_lensing.models.parametric_models import (
    read_parametric_model, params_reader)
from charm_lensing.psf_operator import PsfOperator_fft
from charm_lensing.spaces import Space
from charm_lensing.response import build_integration_operator
from charm_lensing.utils import from_random


def reset_random_seed(seed: int):
    """Reset the random seed for both ift and np."""
    ift.random.push_sseq_from_seed(seed)
    np.random.seed(seed)


def hubble_deep_field_loader(config: dict) -> (np.array, tuple[float]):
    '''Loads an Hubble deep field image'''
    source = np.load(config['hubble_path'], allow_pickle=True).item()
    if config['source_id'] is None:
        source_id = np.random.randint(len(source))
    else:
        source_id = int(config['source_id'])
    source = source[source_id]['image']
    source = source / source.max() * config['source_maximum']
    source[source < 0.0] = 0.0
    return source, (config['distance'],) * 2


def fits_loader(config: dict) -> (np.array, tuple[float]):
    '''Loads an fits image'''
    source = load_fits(config['fits_path'])
    source = source / source.max() * config['source_maximum']
    source[source < 0.0] = 0.0
    return source, (config['distance'],) * 2


class ImageSourceSpline:
    def __init__(self, image: np.array, distances: tuple[float]) -> None:
        source_coordinates_0 = coords(image.shape[0], distances[0])
        source_coordinates_1 = coords(image.shape[1], distances[1])
        self.source_model = RectBivariateSpline(
            source_coordinates_0,
            source_coordinates_1,
            image)

    def brightness_point(self, xycoords: np.array) -> np.array:
        return self.source_model(*xycoords, grid=False)


class ImageSourceBilinear:
    def __init__(self, image: np.array, distances: tuple[float]) -> None:
        self.source_space = Space(image.shape, distances)
        self.source = image

    def brightness_point(self, xycoords: np.array) -> np.array:
        model = build_linear_interpolation(
            ('source', self.source_space),
            ('lens', xycoords.reshape(2, -1).shape)
        )
        return model(
            {'source': self.source, 'lens': xycoords.reshape(2, -1)}
        ).reshape(xycoords.shape[1:])


def get_lens_light(lens_space_model: SpaceModel, mock_config: dict, seed: int):
    """ Returns the lens light and the random seed used to generate it.
    The lens light is defined on the extended space, and has units of flux.
    """

    if mock_config['lens_light'] is not None and mock_config['lens_light'].lower() in [
        'self_consistent']:

        # NOTE: Lens light model is defined in flux units
        x_ll = from_random(
            lens_space_model.light_model.model.domain, seed=seed)
        lens_light = lens_space_model.light_model.model(x_ll)

        # NOTE: Lens light is defined on the extended space
        slc_l = get_centered_slice(
            lens_space_model.space.shape, lens_space_model.space.shape)

        return lens_light[slc_l], x_ll
    else:
        return 0, None


def get_lens_mass(lens_space_model: SpaceModel, mock_config: dict, seed: int):
    """Handles the lens type based on the mock configuration."""
    if mock_config['lens_type'].lower() == 'self_consistent':
        x_ld = from_random(
            lens_space_model.convergence_model.deflection().domain, seed=seed)
        lens_convergence = lens_space_model.convergence_model(x_ld)
        lens_deflection = lens_space_model.convergence_model.deflection()(x_ld)
        lens_deflection = lens_deflection.reshape(2, *lens_convergence.shape)
        slc_d = get_centered_slice(
            lens_space_model.space.extend().shape,
            lens_space_model.space.shape,
            coordinates=True)
        slc = get_centered_slice(
            lens_space_model.space.extend().shape,
            lens_space_model.space.shape,
            coordinates=False)
        return lens_convergence[slc], lens_deflection[slc_d], x_ld

    elif mock_config['lens_type'].lower() in ['direct_set', 'direct']:
        direct = mock_config['direct_set']
        mean_models = [
            (read_parametric_model(model_name.lower()),
             params_reader(read_parametric_model(model_name.lower()), model_params))
            for model_name, model_params in direct.items()
        ]
        lens_convergence = np.zeros(lens_space_model.space.shape)
        lens_deflection = np.zeros((2, *lens_space_model.space.shape))
        for model, params in mean_models:
            assert hasattr(model, 'deflection'), \
                f"""Model {model} does not have a deflection field.
            Direct mock set needs to be adjusted"""
            # TODO: Add support for all models

            lens_convergence += model(params, lens_space_model.space.coords())
            lens_deflection += model.deflection(
                params, lens_space_model.space.coords())

        return lens_convergence, lens_deflection, None


def get_source_light_and_projection(
        source_space_model: SpaceModel,
        lens_space_model: float,
        mock_config: dict,
        seed: int,
        y: ArrayLike
):
    """Creates the source light based on the mock configuration."""

    if mock_config['source_light'].lower() in ['self_consistent']:
        x_sl = from_random(
            source_space_model.light_model.model.domain, seed=seed)
        s = source_space_model.light_model(x_sl)
        s_brightness = s / np.prod(source_space_model.space.distances)
        source_evaluater = build_linear_interpolation(
            ('source', source_space_model.space),
            ('points', y.reshape(2, -1).shape)
        )

        tmp_field = ift.makeField(
            source_evaluater.domain,
            {'source': s_brightness, 'points': y.reshape(2, -1)}
        )
        ls = source_evaluater(tmp_field).val
        Ls = ls.reshape(lens_space_model.space.shape) * \
             lens_space_model.space.nifty_domain.dvol

        return s.T, Ls, x_sl  # FIXME: Why is the transpose needed?

    elif mock_config['source_light'].lower() in ['hubble', 'fits']:
        if mock_config['source_light'].lower() == 'hubble':
            source, distances = hubble_deep_field_loader(mock_config['hubble'])
        elif mock_config['source_light'].lower() == 'fits':
            source, distances = fits_loader(mock_config['fits'])

        source_dvol = np.prod(distances)
        source = source / source_dvol
        if mock_config.get('interpolation', 'bilinear').lower() == 'bilinear':
            source_evaluater = ImageSourceBilinear(source, distances)
            print("Using bilinear interpolation")
        elif mock_config.get('interpolation', 'bilinear').lower() == 'spline':
            source_evaluater = ImageSourceSpline(source, distances)
            print("Using spline interpolation")
        else:
            raise ValueError(
                f"Invalid interpolation method: {mock_config.get('interpolation', 'bilinear')}"
            )
        s = source_evaluater.brightness_point(
            source_space_model.space.coords()
        ) * source_space_model.space.nifty_domain.dvol
        s = s.clip(1e-16)
        Ls = source_evaluater.brightness_point(
            y
        ) * lens_space_model.space.nifty_domain.dvol
        return s, Ls, None

    else:
        raise ValueError("Invalid source light configuration")


def create_mock_data(
        lens_space_model: SpaceModel,
        source_space_model: SpaceModel,
        mock_config: dict,
        psf: Optional[ArrayLike] = None,
        seed: int = 42
) -> tuple[Any, int | float | Any, Any, Any]:

    lens_light, x_ll = get_lens_light(lens_space_model, mock_config, seed)

    lens_convergence, lens_deflection, x_ld = get_lens_mass(
        lens_space_model, mock_config, seed)

    y = lens_space_model.space.coords() - lens_deflection
    s, Ls, x_sl = get_source_light_and_projection(
        source_space_model, lens_space_model, mock_config, seed, y)

    # NOTE: Get mock position unless for first_config which is complete None
    mock_pos_list = [x for x in [x_ll, x_ld, x_sl] if x is not None]
    if len(mock_pos_list) > 0:
        mock_pos = mock_pos_list[0]
        for pos in mock_pos_list[1:]:
            mock_pos.update(pos)

        # from functools import reduce
        # mock_pos = reduce(lambda x, y: x.update(y), mock_pos_list)
    else:
        mock_pos = None

    if psf is None:
        lensed_light = Ls
        light_distribution = Ls + lens_light
    else:
        lensed_light = PsfOperator_fft(Ls, psf)
        light_distribution = PsfOperator_fft(Ls + lens_light, psf)

    # Data space can be different from lens space
    data_space = Space(
        mock_config['data_space']['Npix'],
        mock_config['data_space']['distance'],
        space_key='data'
    )
    integration = build_integration_operator(
        lens_space_model.space,
        data_space.nifty_domain.distances)
    lensed_light = integration(lensed_light)
    light_distribution = integration(light_distribution)

    # NOTE: Set noise scale only for the lensed light
    if mock_config.get('SNR', None) is not None:
        noise_scale = noise_scale_for_target_snr(
            lensed_light, mock_config.get('SNR'),
            mock_config.get('mask_criterion'))
    else:
        noise_scale = mock_config.get('noise_scale')

    reset_random_seed(seed)
    d = light_distribution + np.random.normal(0, noise_scale, data_space.shape)

    return s, d, lens_convergence, lens_deflection, mock_pos, noise_scale, Ls


def func_to_optimize(noise_scale: float, signal: np.ndarray, target_snr: float,
                     mask_criterion: float) -> float:
    estimated_snr = get_average_signal_to_noise(signal, noise_scale,
                                                mask_criterion=mask_criterion)
    return estimated_snr - target_snr


def noise_scale_for_target_snr(
        data: np.ndarray,
        target_snr: float,
        mask_criterion: float,
        fun: Callable = func_to_optimize
) -> np.ndarray:
    # initial guess for the bounds of noise_scale
    lower_bound = 0.001
    upper_bound = 0.9 * np.sqrt(data.max()/mask_criterion)
    noise_scale = brentq(fun, lower_bound, upper_bound, args=(data, target_snr, mask_criterion))
    return noise_scale


def get_average_signal_to_noise(signal: np.ndarray, noise_scale: float,
                                mask_criterion: float) -> float:
    """
    Calculate the signal-to-noise ratio for the given signal.

    Args:
        signal (np.ndarray): The input signal as a NumPy array.
        noise_scale (float): The scale of the noise.
        mask_criterion (float): The criterion for masking the signal.
        Sets the minimum signal (with respect to the noise) to be
        considered when calculating the average signal-to-noise ratio.

    Returns:
        float: The calculated signal-to-noise ratio.

    Raises:
        ValueError: If no signal is above the mask criterion.
    """
    snr = signal / noise_scale
    mask = snr > mask_criterion * noise_scale
    if mask.sum() == 0:
        raise ValueError("No signal above the mask criterion.")
    return snr[mask].mean()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import jax.numpy as jnp

from typing import Tuple
from numpy.typing import ArrayLike


class Gaussian:
    # FIXME : Put in I0
    prior_keys = (('center', (2,)), ('covariance', (2,)),
                  ('off_diagonal', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        center, covariance, off_diagonal = params
        sx, sy = covariance
        theta = off_diagonal

        x, y = coords[0] - center[0], coords[1] - center[1]

        a = jnp.cos(theta)**2/(2*sx**2) + jnp.sin(theta)**2/(2*sy**2)
        b = -jnp.sin(2*theta)/(4*sx**2) + jnp.sin(2*theta)/(4*sy**2)
        c = jnp.sin(theta)**2/(2*sx**2) + jnp.cos(theta)**2/(2*sy**2)

        return jnp.exp(-(a*x**2 + 2*b*x*y + c*y**2))

    # def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
    #     center, covariance, off_diagonal = params
    #     g00, g11 = covariance
    #     g01 = g00*g11*off_diagonal
    #     x, y = coords[0]-center[0], coords[1]-center[1]
    #     det = jnp.abs(g00*g11-g01*g01)
    #     return jnp.exp(-0.5/det * (x**2*g11-x*y*(g01+g01)+y**2*g00))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .rotation import rotation

import jax.numpy as jnp

from typing import Tuple
from numpy.typing import ArrayLike


def piemd_deflection(f, bc, x1, x2):
    fprim = jnp.sqrt(1-f**2)
    b = jnp.sqrt(x1**2 + f**2*x2**2)
    bdx = x1 + 1j * f**2 * x2
    tmp = jnp.sqrt(f)/fprim * (
        jnp.arctanh(fprim*jnp.sqrt(b**2+bc**2)/bdx) -
        jnp.arctanh(fprim*bc/(f*(x1 + 1j*x2)))
    )
    return jnp.array((tmp.real, -tmp.imag))


class Piemd:
    prior_keys = (('b', (1,)), ('rs', (1,)), ('center', (2,)),
                  ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        f, bc = 1/q, rs/q
        return b*jnp.sqrt(f)/(2*jnp.sqrt(x**2+f**2*y**2+bc**2))

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        f, bc = 1/q, rs/q
        return b*jnp.array(rotation(piemd_deflection(f, bc, x, y), -theta))


class Dpie:
    prior_keys = (('b', (1,)), ('rs', (1,)), ('rc', (1,)),
                  ('center', (2,)), ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, rc, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        f, bc_core, bc_cut = 1/q, rs/q, rc/q
        return b*jnp.sqrt(f)*(
            1/(2*jnp.sqrt(x**2 + f**2*y**2 + bc_core**2)) -
            1/(2*jnp.sqrt(x**2 + f**2*y**2 + bc_cut**2))
        )

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, rc, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        f, bc_core, bc_cut = 1/q, rs/q, rc/q
        return b*jnp.array(
            rotation(
                piemd_deflection(f, bc_core, x, y) -
                piemd_deflection(f, bc_cut, x, y),
                -theta))


class Pjaffe:
    prior_keys = (('b', (1,)), ('rs', (1,)), ('center', (2,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, center = params
        R = jnp.hypot(coords[0]-center[0], coords[1]-center[1])

        return b/2 * (1/R - 1/jnp.sqrt(R**2 + rs**2))

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, rs, center = params
        x, y = coords[0]-center[0], coords[1]-center[1]
        return b*jnp.array(
            piemd_deflection(0.9999, 0.0, x, y) -
            piemd_deflection(0.9999, rs, x, y),
        )



#!/usr/bin/env python3
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import nifty8.re as jft
import jax.numpy as jnp

from .parametric_model import ParametricModel
from typing import Callable, Tuple


def shape_adjust(val, shape):
    if jnp.shape(val) == shape:
        return jnp.array(val)
    else:
        return jnp.full(shape, val)


def ducktape(op, key):
    def keyed_call(x):
        return op(x[key])
    return keyed_call


def build_prior_operator(
        domain_key: str, values: dict, shape: Tuple[int] = (1,)
) -> Callable:

    DISTRIBUTION_MAPPING = {
        'normal': (jft.normal_prior, ['mean', 'sigma']),
        'log_normal': (jft.lognormal_prior, ['mean', 'sigma']),
        'lognormal': (jft.lognormal_prior, ['mean', 'sigma']),
        'uniform': (jft.uniform_prior, ['min', 'max']),
        None: (lambda x: lambda _: x, ['mean'])  # FIXME: Is this correct???
    }

    distribution = values.get('distribution')
    transformation = values.get('transformation', None)

    try:
        prior_function, required_params = DISTRIBUTION_MAPPING[distribution]
        vals = [shape_adjust(values[key], shape) for key in required_params]

    except KeyError as e:
        if distribution not in DISTRIBUTION_MAPPING:
            raise NotImplementedError(
                f"{domain_key}: Prior distribution '{distribution}' is not "
                "implemented. Available distributions: \n"
                f"{list(DISTRIBUTION_MAPPING.keys())}"
            ) from e
        else:
            raise KeyError(
                f"{domain_key}: The distribution '{distribution}' requires the"
                f" keys: {required_params}"
            ) from e

    operator = ducktape(prior_function(*vals), domain_key)

    if transformation is not None:
        trafo = getattr(jnp, transformation)
        return lambda x: trafo(operator(x))

    return operator


def check_prior_keys(model: ParametricModel, prior_dict: dict):
    es = '''Not all required keys are specified in the prior dictionary.
    (required) prior keys: {prior_keys}
    (provided) prior dict: {prior_dict}'''
    for key, _ in model.prior_keys:
        if key not in prior_dict:
            raise ValueError(
                es.format(prior_keys=[key[0] for key in model.prior_keys],
                          prior_dict=prior_dict.keys()))


def build_parametric_prior(
        model: ParametricModel, prefix: str, prior_dict: dict
) -> jft.Model:
    check_prior_keys(model, prior_dict)

    ptree = {}
    funcs = []
    for key, shape in model.prior_keys:
        domain_key = '_'.join((prefix, key))
        ptree[domain_key] = jft.ShapeWithDtype((shape))
        op = build_prior_operator(domain_key, prior_dict[key], shape)
        funcs.append(op)

    if len(funcs) == 1:
        prior_model = funcs[0]
    else:
        def prior_model(x):
            return [op(x) for op in funcs]

    return jft.Model(
        prior_model,
        domain=ptree
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from numpy.typing import ArrayLike
import nifty8 as ift

from nifty8.re import Model, Vector
from typing import Union

MATERN_KEYS = ['matern', 'mattern', 'matern_field', 'mkf', 'mf']
ZERO_FLUX_KEYS = ['zero_flux', 'zf']


def update_distribution(dictionary):
    if isinstance(dictionary, dict):
        if ('distribution' in dictionary and dictionary['distribution'] == 'uniform'):
            if 'mean' in dictionary:
                dictionary['min'] = float(dictionary.pop('mean'))
            if 'sigma' in dictionary:
                dictionary['max'] = dictionary['min'] + \
                    float(dictionary.pop('sigma'))
        else:
            for key, value in dictionary.items():
                if isinstance(value, str):
                    try:
                        dictionary[key] = float(value)
                    except ValueError:
                        pass
                elif isinstance(value, list):
                    dictionary[key] = [float(item) if isinstance(
                        item, str) else item for item in value]
                elif isinstance(value, dict):
                    update_distribution(value)

                if key in MATERN_KEYS:
                    dictionary[key]['fluctuations']['renormalize_amplitude'] = False
                    dictionary[key]['fluctuations']['non_parametric_kind'] = 'amplitude'

                if key in ZERO_FLUX_KEYS:
                    out = dictionary[key]
                    dictionary[key] = {'mean': out}


def update_position(position, key_missing_update=False):
    from numpy import array
    key_rename = {
        'lens_ext_lens_ext_deflection_shear_strength': 'lens_ext_convergence_shear_strength',
        'lens_ext_lens_ext_deflection_shear_theta': 'lens_ext_convergence_shear_theta',
    }
    key_rename_2 = {
        'zero_flux': 'zero_flux_mean'
    }
    key_missing = {
        'lens_ext_convergence_shear_center': array([0., 0.]),
    }

    jft_position = {}
    for key in position.keys():
        if key in key_rename:
            jft_position[key_rename[key]] = position[key]
        elif '_'.join(key.split('_')[-2:]) in key_rename_2:
            out = '_'.join(
                key.split('_')[:-2] + [key_rename_2['_'.join(key.split('_')[-2:])]])
            jft_position[out] = position[key]
        else:
            jft_position[key] = position[key]

    if key_missing_update:
        for key in key_missing:
            jft_position[key] = key_missing[key]

    return jft_position


def jft_samples_converter(
        mean: Union[dict, ift.MultiField], samples: ift.SampleList):
    from jax.numpy import array
    from nifty8.re import Samples, Vector

    if isinstance(mean, ift.MultiField):
        mean = mean.val

    out = {key: [] for key in update_position(next(samples.iterator())).keys()}
    for s in samples.iterator():
        s = update_position(s.val)
        for key, value in s.items():
            out[key].append(value - mean[key])

    for key, val in out.items():
        out[key] = array(val)

    return Samples(pos=Vector(mean), samples=Vector(out))


def get_nifty_domain(domain: dict):
    if isinstance(domain, Vector):
        domain = domain.tree
    elif not isinstance(domain, dict):
        return ift.UnstructuredDomain(domain.shape)

    out = {}
    for key, val in domain.items():
        out[key] = ift.UnstructuredDomain(val.shape)
    return ift.makeDomain(out)


def jax_operator(op: Model) -> ift.JaxOperator:
    return ift.JaxOperator(
        domain=get_nifty_domain(op.domain),
        target=get_nifty_domain(op.target),
        func=op.__call__)


def build_gaussian_likelihood(
        domain: ift.UnstructuredDomain,
        data: ArrayLike,
        std: ArrayLike
):
    assert domain.shape == data.shape == std.shape

    v = ift.makeField(domain, std**2)
    d = ift.makeField(domain, data)

    N = ift.DiagonalOperator(v, sampling_dtype=float)
    return ift.GaussianEnergy(data=d, inverse_covariance=N.inverse)


class Passer(ift.LinearOperator):
    def __init__(self, domain, target):
        self._domain = domain
        self._target = target
        self._capability = self.TIMES | self.ADJOINT_TIMES

    def apply(self, x, mode):
        self._check_input(x, mode)
        if mode == self.TIMES:
            out = {t: x.val[t] for t in self.target.keys()}
            return ift.makeField(self.target, out)
        out = {t: x.val[t]
               if t in self.target.keys() else 0 for t in self.domain.keys()}
        return ift.makeField(self.domain, out)


def connect_likelihood_to_model(
        likelihood: ift.LikelihoodEnergyOperator, model: Model
) -> ift.LikelihoodEnergyOperator:
    '''Connect a likelihood to a model.'''

    if not isinstance(model, Model):
        raise ValueError('model must be an instance of nifty8.re.Model')
    model = jax_operator(model)

    ldom = likelihood.domain
    if isinstance(ldom, Vector):
        ldom = ldom.tree

    tdom = {t: ldom[t] for t in ldom.keys() if t not in model.target.keys()}
    mdom = {t: model.domain[t] for t in model.domain.keys()}
    mdom.update(tdom)
    passer = Passer(ift.makeDomain(mdom), ift.makeDomain(tdom))

    return likelihood @ (model + passer)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import nifty8.re as jft
from charm_lensing import (spaces, space_model, utils, projection_operator)

from dataclasses import dataclass


@ dataclass
class LensSystem:
    lens_plane_model: space_model.SpaceModel
    source_plane_model: space_model.SpaceModel
    source_projector: jft.Model

    def _get_forward_model(self, only_source=False, parametric=False):
        deflection_model = self.lens_plane_model.convergence_model.deflection(
            parametric=parametric)

        ptree = utils.chain_model_domains(
            deflection_model.domain,
            self.source_plane_model.light_model.domain
        )

        forward = jft.Model(
            lambda x: {'sky': self.source_projector(
                {'source': self.source_plane_model.light_model(x),
                 'deflection': deflection_model(x)}
            )},
            domain=ptree)

        lens_light = self.get_lens_light()
        if lens_light is None or only_source:
            return forward
        else:
            ptree = utils.chain_model_domains(ptree, lens_light.domain)
            return jft.Model(
                lambda x: {'sky': forward(x)['sky'] + lens_light(x)},
                domain=ptree
            )

    def get_forward_model_full(self, only_source=False):
        '''Return the forward pass of the nonparmetric or hybrid model to
        produce the sky emission.

        Parameters
        ----------
        only_source : bool
            If true, the lens light model is not included in the forward model.
        '''
        return self._get_forward_model(only_source=only_source, parametric=False)

    def get_forward_model_parametric(self, only_source=False):
        '''Return the forward pass of the parametric model to
        produce the sky emission.

        Parameters
        ----------
        only_source : bool
            If true, the lens light model is not included in the forward model.
        '''
        return self._get_forward_model(only_source=only_source, parametric=True)

    def get_lens_light(self):
        if self.lens_plane_model.light_model is None:
            return None
        return self.lens_plane_model.light_model

    def critical_caustic(self, position: dict, plot=False):
        '''Calculate the critical and caustic lines of the lens system.

        Parameters
        ----------
        position : dict
            A position in latent space of the lens system.
        plot : bool
            If true, the critical and caustic are plotted.
        '''
        from .critical_and_caustic import calculate_critical_caustic as clcc
        return clcc(self, position, plot=plot)


def build_lens_system(cfg: dict) -> LensSystem:
    '''
    Parse the configuration file and build the lens system and the projection 
    operator.

    Parameters
    ----------
    cfg : dict, The configuration file which has to hold
        - 'lens_space' : dict
          - 'Npix' : int
          - 'distance' : float
          - 'padding_ratio' : float (default=0.0)
        - 'source_space' : dict
          - 'Npix' : int
          - 'distance' : float
          - 'padding_ratio' : float (default=0.0)
        - 'model' : dict
          - 'lens' : dict (The lens model)
          - 'source' : dict (The source light model)
    '''

    # Spaces and larger Spaces to avoid periodic boundary conditions
    lens_space = spaces.Space(
        cfg['spaces']['lens_space']['Npix'],
        cfg['spaces']['lens_space']['distance'],
        'lens',
        cfg['spaces']['lens_space']['padding_ratio'])

    source_space = spaces.Space(
        cfg['spaces']['source_space']['Npix'],
        cfg['spaces']['source_space']['distance'],
        'source',
        cfg['spaces']['source_space']['padding_ratio'])

    lens_plane_model = space_model.build_space_model(
        lens_space,
        cfg['model']['lens'],
        light_extend=False
    )

    source_plane_model = space_model.build_space_model(
        source_space,
        cfg['model']['source'],
        light_extend=True
    )

    source_projector = projection_operator.build_projection_operator(
        lens_space, source_space, cfg['spaces'].get('interpolation', 'bilinear'))

    return LensSystem(
        lens_plane_model=lens_plane_model,
        source_plane_model=source_plane_model,
        source_projector=source_projector
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from charm_lensing.spaces import Space

import nifty8.re as jft

import jax
import jax.numpy as jnp

from typing import Tuple, Callable
from functools import partial


def build_xy_jnp(dom_shape, dom_dist, sampling_points):
    """Builds the jax array for the xy coordinates."""
    N_dim, N_points = sampling_points.shape

    mg = jnp.mgrid[(slice(0, 2),) * N_dim]
    mg = jnp.array(list(map(jnp.ravel, mg)))
    dist = jnp.array(dom_dist).reshape(-1, 1)
    pos = sampling_points / dist
    excess = pos - jnp.floor(pos)
    pos = jnp.floor(pos).astype(jnp.int64)
    max_index = jnp.array(dom_shape).reshape(-1, 1)

    outside = jnp.any((pos > max_index) + (pos < 0), axis=0)

    data = jnp.zeros((2, len(mg[0]), N_points))
    jj = jnp.zeros((len(mg[0]), N_points), dtype=jnp.int64)

    for i in range(len(mg[0])):
        quadrant = jnp.abs(1 - mg[:, i].reshape(-1, 1) - excess)
        data = data.at[:, i, :].set(quadrant)

        data = jnp.where(outside, 0., data)

        fromi = (pos + mg[:, i].reshape(-1, 1)) % max_index
        jj = jj.at[i, :].set(fromi[0] * dom_shape[0] + fromi[1])

    return jj, data


def forward(ff, val):
    return jnp.sum(ff * val, axis=0)


vmap_forward = jax.vmap(forward, in_axes=(None, 0))


def build_interpolation(
        build_xy: Callable,
        dom_dist: Tuple[float]
) -> Callable:

    @jax.custom_jvp
    def interpolation(position, field):
        jj, quadrants = build_xy(position)
        return forward(field[jj], jnp.prod(quadrants, axis=0))

    @interpolation.defjvp
    def interpolation_jvp(primals, tangents):
        position, field = primals
        position_dot, field_dot = tangents

        jj, quadrants = build_xy(position)
        primal_val = jnp.prod(quadrants, axis=0)
        grad_x_val = quadrants[1] * \
            jnp.array((-1, -1, 1, 1))[..., None] / dom_dist[0]
        grad_y_val = quadrants[0] * \
            jnp.array((-1, 1, -1, 1))[..., None] / dom_dist[1]

        # position
        primal_out, grad_x, grad_y = vmap_forward(
            field[jj], jnp.array((primal_val, grad_x_val, grad_y_val))
        )

        # field
        field_dot = forward(field_dot[jj], primal_val)

        tangent_out = (
            jnp.sum(jnp.array((grad_x, grad_y)) * position_dot, axis=0) +
            field_dot
        )

        return primal_out, tangent_out

    return interpolation


def build_linear_interpolation(
        regular_grid: Tuple[str, Space],
        points: Tuple[str, Tuple[int, int]]
):
    """
    Multilinear interpolation for variable points in an RGSpace

    Parameters
    ----------
    regular_grid: Tuple[str, Space]
        Tuple containing the domain key and the regular grid, 2D space,
        on which to interpolate.
    points: Tuple[str, Tuple[int, int]]
        Tuple containing the domain key of the points to be interpolated
        and their shape.
    """

    rgrid_key, rgrid_dom = regular_grid
    points_key, points_shape = points

    # Checks and balances
    if not isinstance(rgrid_dom, Space):
        raise TypeError
    if len(rgrid_dom.shape) != 2:
        raise TypeError
    if len(points_shape) != 2:
        raise ValueError('Point domain shape length incompatible')
    if points_shape[0] != len(rgrid_dom.shape):
        raise ValueError('Point domain incompatible with RGSpace')

    parameter_tree = {
        rgrid_key: jft.ShapeWithDtype(rgrid_dom.shape),
        points_key: jft.ShapeWithDtype(points_shape)
    }

    # CHECKED: This is doing the correct thing:
    # upper_left is the coordiante of the center of the upper left pixel of
    # the rgrid domain.
    # Consequently, the (interpolated) points which are centered with respect
    # to the image center get shifted in the interpolation by upper_left.
    # This conforms to the convention that the interpolation assumes that the
    # upper left corner of the image is at (0,0).
    upper_left = rgrid_dom.coords()[..., 0, 0]
    upper_left = upper_left[..., None]

    build_xy = partial(
        build_xy_jnp,
        rgrid_dom.shape,
        rgrid_dom.distances)
    inter = jax.jit(build_interpolation(
        build_xy,
        rgrid_dom.distances))

    def interpolation(x):
        return inter(
            x[points_key] - upper_left,
            x[rgrid_key].reshape(-1)
        )

    return jft.Model(
        interpolation,
        domain=parameter_tree
    )


def build_finufft_interpolation(
    regular_grid: Tuple[str, Space],
    points: Tuple[str, Tuple[int, int]]
):
    """
    Build interpolation based on the jax_finufft library for variable points
    in RGSpace

    Parameters
    ----------
    regular_grid: Tuple[str, Space]
        Tuple containing the domain key and the regular grid, 2D space,
        on which to interpolate.
    points: Tuple[str, Tuple[int, int]]
        Tuple containing the domain key of the points to be interpolated
        and their shape.
    """

    rgrid_key, rgrid_dom = regular_grid
    points_key, points_shape = points

    # Checks and balances
    if not isinstance(rgrid_dom, Space):
        raise TypeError
    if len(rgrid_dom.shape) != 2:
        raise TypeError
    if len(points_shape) != 2:
        raise ValueError('Point domain shape length incompatible')
    if points_shape[0] != len(rgrid_dom.shape):
        raise ValueError('Point domain incompatible with RGSpace')

    if rgrid_dom.distances[0] != rgrid_dom.distances[1]:
        raise ValueError('Finufft only supports uniform grids')
    if rgrid_dom.shape[0] != rgrid_dom.shape[1]:
        raise ValueError('Finufft only supports square grids')

    parameter_tree = {
        rgrid_key: jft.ShapeWithDtype(rgrid_dom.shape),
        points_key: jft.ShapeWithDtype(points_shape)
    }

    upper_left = rgrid_dom.coords()[..., 0, 0]
    upper_left = upper_left[..., None]

    rgrid_extent = rgrid_dom.distances[0] * rgrid_dom.shape[0]
    from jax_finufft import nufft2

    # # FIXME: THIS SHOULD BE DONE IN THE MAIN SCRIPT
    # import jax
    # jax.config.update("jax_platform_name", "cpu")

    def finufft_interpolation(pos, field):
        x_finufft = (pos - upper_left) / rgrid_extent * 2 * jnp.pi

        f_field = jnp.fft.ifftshift(jnp.fft.ifft2(field))
        n_field = nufft2(
            f_field, x_finufft[0].reshape(-1), x_finufft[1].reshape(-1)
        )
        return n_field.real

    return jft.Model(
        lambda x: finufft_interpolation(x[points_key], x[rgrid_key]),
        domain=parameter_tree
    )



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .rotation import rotation

import jax.numpy as jnp

from typing import Tuple
from numpy.typing import ArrayLike


def hyp2f1_series(t, q, r, z, max_terms=10):
    """
    Computes the Hypergeometric function numerically
    according to the recipe in O'Riordan III.
    """

    # U from
    q_ = (1 - q ** 2) / (q ** 2)
    u = 0.5 * (1 - jnp.sqrt(1 - q_ * (r / z) ** 2))
    # First coefficient
    a_n = 1.0
    # Storage for sum
    F = jnp.zeros_like(z, dtype='complex64')

    for n in range(max_terms):
        F += a_n * (u ** n)
        a_n *= ((2 * n) + 4 - (2 * t)) / ((2 * n) + 4 - t)
    return F


class Spl:
    prior_keys = (('b', (1,)), ('t', (1,)), ('center', (2,)),
                  ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        '''(2-t)/2 (b/r)**t'''
        b, t, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)
        return ((2 - t) / 2) * (b / jnp.hypot(q*x, y)) ** t

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, t, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        z = x + 1j * y
        er = jnp.hypot(q*x, y)

        f1 = (b ** 2) / (q * z)
        f2 = (b / er) ** (t - 2)
        f3 = hyp2f1_series(t, q, er, z)

        # Rotate back into original frame
        sol = (f1 * f2 * f3).conjugate() * jnp.exp(- 1j * theta)
        return jnp.array((sol.real, sol.imag))


class Bpl:
    prior_keys = (('b', (1,)), ('t1', (1,)), ('t2', (1,)), ('rs', (1,)),
                  ('center', (2,)), ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        kB, t1, t2, rB, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        # Elliptical radius / break radius
        x = jnp.hypot(x*q, y) / rB

        c1 = jnp.where((x < 1), x**(-t1), 0)
        c2 = jnp.where((x >= 1), x**(-t2), 0)
        return kB*(c1 + c2)

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        kB, t1, t2, rB, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        z = (x + 1j * y) * jnp.exp(+1j*theta)

        R = jnp.hypot(z.real*q, z.imag)

        # Factors common to eq. 18 and 19
        factors = 2 * kB * (rB ** 2) / (q * z * (2 - t1))

        # Hypergeometric functions
        F1 = hyp2f1_series(t1, q, R, z, max_terms=10)
        F2 = hyp2f1_series(t1, q, rB, z, max_terms=10)
        F3 = hyp2f1_series(t2, q, R, z, max_terms=10)
        F4 = hyp2f1_series(t2, q, rB, z, max_terms=10)

        x1 = rB/R
        c1 = jnp.where((1/x1 <= 1), F1 * x1**(t1 - 2), 0)
        c2 = jnp.where((1/x1 > 1), (F2 + (2-t1)/(2-t2)
                                    * ((x1**(t2 - 2)) * F3 - F4)), 0)

        sol = (factors * (c1 + c2)).conjugate() * jnp.exp(-1j*theta)
        return jnp.array((sol.real, sol.imag))


class Cpl:
    prior_keys = (('b', (1,)), ('t', (1,)), ('rs', (1,)),
                  ('center', (2,)), ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        b, t, rs, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        er = jnp.hypot(q*x, y)
        return b * ((2 - t) / 2) / ((er**2 + rs**2)**t)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import jax.numpy as jnp

from typing import Tuple
from numpy.typing import ArrayLike


class Shear:
    prior_keys = (('strength', (1,)), ('theta', (1,)), ('center', (2,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ss, sa, center = params
        xx, yy = coords - center.reshape(2, -1)
        ax = jnp.cos(2*sa)*xx + jnp.sin(2*sa)*yy
        ay = jnp.sin(2*sa)*xx - jnp.cos(2*sa)*yy
        return ss*jnp.array((ax, ay))

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ss, sa, center = params
        xx, yy = coords - center.reshape(2, -1)
        ax = jnp.cos(2*sa)*xx + jnp.sin(2*sa)*yy
        ay = jnp.sin(2*sa)*xx - jnp.cos(2*sa)*yy
        return ss*jnp.array((ax, ay))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import jax.numpy as jnp

from numpy.typing import ArrayLike
from typing import Tuple


class Constant:
    prior_keys = (('value', (1,)),)

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        return jnp.full_like(coords[0], params[0])


class ZeroFlux:
    prior_keys = (('mean', (1,)),)

    def __call__(self, params: Tuple[float]) -> ArrayLike:
        return params[0]


class LinearShift:
    prior_keys = (('ax', (1,)), ('ay', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ax, ay = params
        out0 = jnp.full_like(coords[0], ax)
        out1 = jnp.full_like(coords[1], ay)
        return jnp.array((out0, out1))

    def deflection(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ax, ay = params
        out0 = jnp.full_like(coords[0], ax)
        out1 = jnp.full_like(coords[1], ay)
        return jnp.array((out0, out1))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


from .rotation import rotation

import jax.numpy as jnp

from typing import Tuple
from numpy.typing import ArrayLike


class ExponentialDisk:
    prior_keys = (('ie', (1,)), ('re', (1,)), ('center', (2,)),
                  ('theta', (1,)), ('q', (1,)))

    def __call__(self, params: Tuple[float], coords: ArrayLike) -> ArrayLike:
        ie, re, center, theta, q = params
        x, y = rotation((coords[0]-center[0], coords[1]-center[1]), theta)

        r = jnp.sqrt(q**2 * x**2 + y**2)
        return ie * jnp.exp(-r / re)



#!/usr/bin/env python3
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import numpy as np
from nifty8 import RGSpace
from numpy import isscalar, empty, ndarray
from numpy.typing import ArrayLike

from dataclasses import dataclass, field
from typing import Tuple, Callable


def central_mask_operator(domain, target):
    sl = []
    for i in range(len(domain.shape)):
        slStart = int((domain.shape[i] - target.shape[i]) / 2.)
        slStop = slStart + target.shape[i]
        sl.append(slice(slStart, slStop, 1))
    slices = tuple(sl)

    def central_mask(x):
        return x[slices]

    return central_mask


def coords(shape: int, distance: float) -> ArrayLike:
    '''Returns coordinates such that the edge of the array is
    shape/2*distance'''
    halfside = shape/2 * distance
    return np.linspace(-halfside+distance/2, halfside-distance/2, shape)


def get_extent(
        shape: tuple[int],
        distances: tuple[float],
        center: tuple[float] = (0, 0)
) -> tuple[float]:
    assert len(shape) == 2

    if isscalar(distances):
        distances = (float(distances),) * len(shape)
    halfside = np.array(shape)/2 * np.array(distances)

    return ((-halfside[0]+center[0], halfside[0]+center[0]) +
            (-halfside[1]+center[1], halfside[1]+center[1]))


def get_xycoords(shape: tuple[int], distances: tuple[float]) -> ArrayLike:
    assert len(shape) == 2
    if isscalar(distances):
        distances = (float(distances),) * len(shape)
    x_direction = coords(shape[0], distances[0])
    y_direction = coords(shape[1], distances[1])
    return np.array(np.meshgrid(x_direction, y_direction, indexing='xy'))


def get_klcoords(shape: tuple[int], distances: tuple[float]) -> ArrayLike:
    assert len(shape) == 2
    if isscalar(distances):
        distances = (float(distances),) * len(shape)
    kx = [np.fft.fftfreq(shape[i], d=distances[i]) for i in range(len(shape))]
    return np.array(np.meshgrid(kx[1], kx[0]))


def get_centered_slice(
    extended_shape: Tuple[int],
    original_shape: Tuple[int],
    coordinates: bool = False
) -> Tuple[slice]:
    slices = []

    for oshp_dim, eshp_dim in zip(original_shape, extended_shape):
        center = eshp_dim // 2
        half_side = oshp_dim // 2
        start_idx = center - half_side
        end_idx = start_idx + oshp_dim

        # Ensure the indices are within the extended array shape
        start_idx = max(0, min(start_idx, eshp_dim - 1))
        end_idx = max(0, min(end_idx, eshp_dim))

        slices.append(slice(start_idx, end_idx))

    if coordinates:
        slices.insert(0, slice(None))

    return tuple(slices)


@ dataclass
class Space:
    """Represents a regular 2D Cartesian grid.

    Parameters
    ----------
    shape : tuple of int
    distances : tuple of float
    """
    shape: Tuple[int]
    distances: Tuple[float]
    space_key: str = ''
    extend_factor: float = 1.0
    center: ndarray = field(default_factory=lambda: np.array((0., 0.)))
    extent: Tuple[float] = field(init=False)
    nifty_domain: RGSpace = field(init=False)

    def __post_init__(self):
        self.shape = tuple(int(s) for s in self.shape)
        if isscalar(self.distances):
            self.distances = (float(self.distances),) * len(self.shape)
        else:
            tmp = empty(len(self.shape), dtype=float)
            tmp[:] = self.distances
            self.distances = tuple(tmp)

        self.center = np.array(self.center)
        self.extent = get_extent(self.shape, self.distances, self.center)
        self.nifty_domain = RGSpace(self.shape, self.distances)

    def coords(self) -> ndarray:
        return get_xycoords(self.shape, self.distances) + self.center[..., None, None]

    def extend(self, extend=None) -> 'Space':
        """Returns a new Space with the shape extended by the extend_factor."""
        if extend is None:
            extend = self.extend_factor
        extended_shape = tuple(int(s * extend) for s in self.shape)
        return Space(
            extended_shape,
            self.distances,
            self.space_key + '_ext', 1.0,
            self.center
        )

    def get_mask_operator(self) -> Callable[ArrayLike, ArrayLike]:
        return central_mask_operator(
            self.extend().nifty_domain, self.nifty_domain)

    def get_mask(
        self, coordinates: bool = False, as_array: bool = False
    ) -> slice | ArrayLike:

        slice = get_centered_slice(
            self.extend().shape, self.shape, coordinates=coordinates)

        if as_array:
            arr = np.full(
                self.extend().coords().shape if coordinates else
                self.extend().shape,
                False)
            arr[slice] = True
            return arr

        return slice



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig


import numpy as np
import yaml

from os.path import join

import jax.numpy as jnp

from nifty8 import ResidualSampleList
from scipy.stats import multivariate_normal
from astropy.io import fits
from astropy.io.fits.hdu.image import ImageHDU

from charm_lensing.psf_operator import PsfOperator_fft

from numpy.typing import ArrayLike
from typing import Dict, Union

from nifty8.re import ShapeWithDtype, Vector


def chain_model_domains(*model_domains):
    ptree = {}
    for model_keys in model_domains:
        for key, val in model_keys.items():
            if key in ptree:
                raise Warning('Duplicate model key: {}'.format(key))
            ptree.update({key: val})
    return ptree


def full(domain: Union[Dict[str, ShapeWithDtype], Vector], value: float):
    if isinstance(domain, Vector):
        domain = domain.tree

    out = {}
    for key, s in domain.items():
        out.update({key: jnp.full(s.shape, value)})
    return out


def from_random(
    domain: Union[Dict[str, ShapeWithDtype], Vector],
    seed: int = 42
):
    if isinstance(domain, Vector):
        domain = domain.tree

    from jax import random
    from nifty8.re import random_like
    key, subkey = random.split(random.PRNGKey(seed))
    return random_like(subkey, domain)


def unite_dict(a: dict, b: dict) -> dict:
    '''Returns: union of a and b'''
    tmp = {}
    tmp.update(a)
    tmp.update(b)
    return tmp


def save_fits(data, name):
    hdu = fits.PrimaryHDU(data)
    hdul = fits.HDUList([hdu])
    if name.split('.')[-1] == 'fits':
        hdul.writeto(name, overwrite=True)
    else:
        hdul.writeto(name+'.fits', overwrite=True)


def load_fits(path_to_file, fits_number=0, get_header=False):
    with fits.open(path_to_file) as hdul:
        headers = [h.header for h in hdul if isinstance(h, ImageHDU)]
        datas = [h.data for h in hdul if isinstance(h, ImageHDU)]
        if len(headers) == 0:
            header = hdul[0].header
            data = hdul[0].data
        else:
            header = headers[fits_number]
            data = datas[fits_number]
    if get_header:
        return np.array(data).astype(np.float64), header
    return np.array(data).astype(np.float64)


def reconstruction_loader(
        reconstruction_path: str,
):
    '''
    Load mean, samples and necessery objects from reconstruction_path/pickle.
    '''
    # TODO load config also for the mean and samples for nifyt8.re
    mean = ResidualSampleList.load_mean(
        join(reconstruction_path, 'pickle', 'last'))
    samples = ResidualSampleList.load(
        join(reconstruction_path, 'pickle', 'last'))

    # load config
    with open(join(reconstruction_path, 'config.yaml'), 'r') as file:
        cfg = yaml.safe_load(file)

    return mean, samples, cfg


def lognormal_pdf_display(mean, std):
    from nifty8 import LognormalTransform, from_random
    lognorm = LognormalTransform(mean, std, 'test', 1000)
    test = lognorm(from_random(lognorm.domain)).val
    mean, std = test.mean(), test.std()

    from matplotlib.pyplot import subplots, show, title
    fig, ax = subplots()
    ax.hist(test)
    ax.axvline(mean, color='red')
    ax.axvline(mean+std, color='black', linestyle='dashed')
    ax.axvline(mean-std, color='black', linestyle='dashed')
    title(f'mean: {mean:.2f}, std: {std:.2f}')
    show()


smoother = multivariate_normal.pdf(
    np.array(np.meshgrid(*(np.arange(-10, 10, 1),)*2)).T,
    mean=(0, 0),
    cov=2*np.eye(2)
)
smoother = smoother/smoother.sum()


def estimate_snr(data: ArrayLike, noise_scale: float) -> float:
    snr_mask = PsfOperator_fft(data, smoother) > 2 * noise_scale
    return data[snr_mask].sum() / (noise_scale * np.sqrt(snr_mask.sum()))



#!/usr/bin/env python3
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Julian Ruestig, Matteo Guardiani


import nifty8 as ift
import numpy as np
from numpy.typing import ArrayLike

import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.patches import Rectangle
import matplotlib.font_manager as fm
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

from functools import partial

from charm_lensing.build_lens_system import LensSystem
from charm_lensing.analysis_tools import shift_and_resize, source_distortion_ratio
from charm_lensing.models import Model
from charm_lensing.data_loader import MockData, ImageData

from typing import List, Any, Callable, Dict, Optional

import nifty8.re as jft

try:
    from jubik0.library.plot import plot_sample_and_stats
except ImportError:
    plot_sample_and_stats = None

STANDARD_DEVIATIONS = 3
LOG_MIN = 1e-6
MINMAX = 2


def find_best_grid(N):
    """Find the best grid arangement for a given number of panels to plot."""

    from math import ceil, sqrt
    # Initialize variables
    best_X, best_Y, min_overhead = None, None, float('inf')

    # Initial guess for X and Y
    initial_guess = ceil(sqrt(N))

    for X in range(initial_guess, 0, -1):
        Y = ceil(N / X)
        overhead = X * Y - N

        if overhead < min_overhead:
            best_X, best_Y, min_overhead = X, Y, overhead

    return best_X, best_Y


def display_text(ax: plt.Axes, text: Dict, **kwargs):
    '''Display text on plot
    ax: matplotlib axis
    text: dict or str (default: {'s': str, 'color': 'white'})
    kwargs:
    - keyword: str
        options: 'top_left' (default), 'top_right', 'bottom_left', 'bottom_right'
    - x_offset_ticker: float (default: 0)
    - y_offset_ticker: float (default: 0)
    '''
    keyword = kwargs.get('keyword', 'top_left')
    x_offset_ticker = kwargs.get('x_offset_ticker', 0)
    y_offset_ticker = kwargs.get('y_offset_ticker', 0)

    if type(text) is str:
        text = {'s': text, 'color': 'white'}

    if keyword == 'top_left':
        ax.text(x=0.05 + x_offset_ticker*0.05,
                y=0.95 - y_offset_ticker*0.05,
                ha='left',
                va='top',
                transform=ax.transAxes,
                **text)
    elif keyword == 'top_right':
        ax.text(x=0.95 - x_offset_ticker*0.05,
                y=0.95 - y_offset_ticker*0.05,
                ha='right',
                va='top',
                transform=ax.transAxes,
                **text)
    elif keyword == 'bottom_left':
        ax.text(x=0.05 + x_offset_ticker*0.05,
                y=0.05 + y_offset_ticker*0.05,
                ha='left',
                va='bottom',
                transform=ax.transAxes,
                **text)
    elif keyword == 'bottom_right':
        ax.text(x=0.95 - x_offset_ticker*0.05,
                y=0.05 + y_offset_ticker*0.05,
                ha='right',
                va='bottom',
                transform=ax.transAxes,
                **text)
    else:
        raise ValueError(
            "Invalid keyword. Use 'top_left', 'top_right', 'bottom_left', or 'bottom_right'.")


def display_colorbar(ax, im, **kwargs):
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(ax)

    cbar_setting = kwargs.get(
        'cbar_setting', dict(position='right', display=True, label=None))

    if cbar_setting.get('display', False) or kwargs.get('display_cbar', False):
        if cbar_setting['position'] == 'bottom':
            cax = divider.append_axes("bottom", size="5%", pad=0.1)
            orientation = 'horizontal'
            cbar_label = 'bottom'
        elif cbar_setting['position'] == 'top':
            cax = divider.append_axes("top", size="5%", pad=0.1)
            orientation = 'horizontal'
            cbar_label = 'top'
        elif cbar_setting['position'] == 'left':
            cax = divider.append_axes("left", size="5%", pad=0.1)
            orientation = 'vertical'
            cbar_label = 'left'
        elif cbar_setting['position'] == 'right':
            cax = divider.append_axes("right", size="5%", pad=0.1)
            orientation = 'vertical'
            cbar_label = 'right'
        else:
            raise Exception

        bar = plt.colorbar(im, cax=cax, orientation=orientation)
        if orientation == 'horizontal':
            cax.xaxis.set_ticks_position(cbar_label)
            cax.xaxis.set_label_position(cbar_label)
        elif orientation == 'vertical':
            cax.yaxis.set_ticks_position(cbar_label)
            cax.yaxis.set_label_position(cbar_label)
        bar.set_label(cbar_setting.get('label', None))
    else:
        position = cbar_setting.get('position', 'right')
        cax = divider.append_axes(position, size="5%", pad=0.1)
        cax.set_visible(False)


def configer_panel(ax, im, **kwargs):
    '''
    cbar_setting: dict
        display: bool
        label: str
        position: str, options: right (default), left, top, bottom

    scalebar: dict
        display: bool
        size: float
        unit: str (default: '"')

    '''
    # Set Colorbar
    display_colorbar(ax, im, **kwargs)

    # Display image size as scalebar
    scalebar_kwargs = kwargs.get('scalebar', dict(display=False))
    if kwargs.get('display_scalebar', False) or scalebar_kwargs.get('display'):
        size = scalebar_kwargs.get('scale', 1)
        slabel = r'${}{}$'.format(size, scalebar_kwargs.get('unit', '"'))
        font_properties = scalebar_kwargs.get(
            'font_properties', fm.FontProperties(size=8))
        color = scalebar_kwargs.get('color', 'white')
        ax.add_artist(AnchoredSizeBar(
            ax.transData,
            size,
            slabel,
            'lower left',
            pad=0.1,
            color=color,
            frameon=False,
            fontproperties=font_properties,
        ))

    # Set axis limits
    ax.set_xlim(kwargs.get('xlim', None))
    ax.set_ylim(kwargs.get('ylim', None))

    # Display texts
    texts = kwargs.get('text_kwargs', None)
    if isinstance(texts, dict):
        if 's' in kwargs:
            texts['text']['s'] = kwargs['s']
        texts = [texts]
    for text_kwargs in texts:
        display_text(ax, **text_kwargs)

    # Remove ticks and labels
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.set_aspect('equal')


def configure_matplotlib(
        small_size=8,
        medium_size=8,
        big_size=8,
        font='Nimbus Roman',
        usetex=True,
        widths='mnras',  # 'aanda'
):
    '''Configure matplotlib settings.
    Outputs: column_width, text_width, font_properties'''

    font_properties = fm.FontProperties(size=small_size)

    # controls default text sizes
    plt.rc('font', size=small_size, family='Nimbus Roman')
    plt.rc('axes', titlesize=small_size)     # fontsize of the axes title
    plt.rc('axes', labelsize=medium_size)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=small_size)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=small_size)    # fontsize of the tick labels
    plt.rc('legend', fontsize=small_size)    # legend fontsize
    plt.rc('figure', titlesize=big_size)
    plt.rc('text', usetex=True)

    if widths == 'mnras':
        column_width = 244*1/72
        text_width = 508*1/72
    elif widths == 'aanda':
        column_width = 88 / 25.4  # mm to inch
        text_width = 170 / 25.4  # mm to inch

    return column_width, text_width, font_properties


def row_imshow(
        row: List[Any],  # Subplot axes
        data: List[ArrayLike],
        titles: List[str],
        settings: List[dict]
):
    for ax, datum, title, setting in zip(row, data, titles, settings):
        # ax.set_title(title)
        im = ax.imshow(datum, **setting, interpolation='none')
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        ax.tick_params(axis='both', which='both', direction='in')
        ax.text(0.95, 0.05, title, transform=ax.transAxes,
                horizontalalignment='right', verticalalignment='bottom',
                color='k', fontsize=10,
                bbox=dict(facecolor='w', edgecolor='none', pad=1.0, alpha=0.8))
        plt.colorbar(im, cax=cax)


def hist_plot(
        ax: plt.Axes,
        data: ArrayLike,
        model: ArrayLike,
        noise: float = 0,
        s: str = 'Lens light',
        xaxis_label: str = 'Input flux [arbitrary  units]',
        yaxis_label: str = 'Model flux [arbitrary  units]',
        min_flux: float = 0.01
):
    d = data.flatten()[data.flatten() > min_flux]
    m = model.flatten()[data.flatten() > min_flux]
    im = ax.hist2d(d, m, bins=100, norm=LogNorm(), cmap='gnuplot2')
    max, min = np.max([d.max(), m.max()]), np.min([d.min(), m.min()])

    xy = np.arange(min, max, 0.1)
    ax.plot(*(xy,)*2, label='y=x', linestyle='-', color='black',
            linewidth=0.5)
    if noise > 0:
        ax.plot(xy, xy+noise, label='y=x', linestyle='--', color='black',
                linewidth=0.5)
        ax.plot(xy, xy-noise, label='y=x', linestyle='--', color='black',
                linewidth=0.5)

    ax.set_xlabel(xaxis_label)
    ax.set_ylabel(yaxis_label)
    ax.set_xlim(min, max)
    ax.set_ylim(min, max)
    ax.tick_params(axis='both', which='both', direction='in')
    text = dict(
        color='black',
        bbox=dict(facecolor='white', edgecolor='none', pad=1.0, alpha=0.5),
        s=s
    )
    display_text(ax, text)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.1)
    plt.colorbar(im[3], cax=cax)


def parameter_check(model: Model, position: ift.MultiField, ax, color='white'):
    if (model is None) or (position is None):
        return None
    elif (model.parametric() is None):
        return None

    extent = model.space.extend().extent
    top = extent[3]
    left = extent[0]

    _, models, priors = model.parametric(full_info=True)
    itercounter = -1
    for m, prior in zip(models, priors):
        itercounter += 1
        mean_params = prior(position)
        for s, val in zip(m.prior_keys, mean_params):
            itercounter += 1
            symbol = s[0]
            value = ', '.join([f'{v:.2f}' for v in val])
            ax.text(
                left*0.95, top*(0.90 - (itercounter) * 0.1),
                f'{symbol} = {value}', color=color, alpha=0.5,
            )


def add_box(ax, extent, linestyle='-', edgecolor='white'):
    xy = (extent[0], extent[0])
    width = extent[1]-extent[0]
    height = extent[1]-extent[0]
    ax.add_patch(Rectangle(xy=xy,
                           width=width,
                           height=height,
                           linestyle=linestyle,
                           edgecolor=edgecolor,
                           facecolor='none',
                           lw=1))


def mock_data_plot(
        position: Dict[str, ArrayLike],
        output_name_template: Optional[str],
        data: Dict[str, MockData],
        lens_system: LensSystem,
        iteration: int,
        parametric: bool = True,
        log_data: bool = False,
) -> None:

    key = next(iter(data.keys()))
    input_source = data[key].source
    convergence_data = data[key].convergence
    deflection_data = data[key].deflection
    R = data[key].response

    # Calculate the reconstructed source
    source_extent = lens_system.source_plane_model.space.extend().extent
    source_mask = lens_system.source_plane_model.space.get_mask_operator()
    source_mean = source_mask(
        lens_system.source_plane_model.light_model.parametric()(position)
    ).T
    source_perturbations = source_mask(
        lens_system.source_plane_model.light_model.nonparametric()(position)
    ).T
    source_reconstruction = source_mask(
        lens_system.source_plane_model.light_model(position)
    ).T  # FIXME: Why the hell is here a transpose?
    source_reconstruction, sx, sy = shift_and_resize(
        input_source, source_reconstruction, full_info=True)
    source_mean = shift_and_resize(input_source, source_mean, sxsy=(sx, sy))
    source_perturbations = shift_and_resize(
        input_source, source_perturbations, sxsy=(sx, sy))

    lens_light_model = lens_system.get_lens_light()
    deflection_mean = lens_system.lens_plane_model.convergence_model.deflection(
        parametric=True)
    forward_parametric = lens_system.get_forward_model_parametric()

    if parametric:
        convergence_model = lens_system.lens_plane_model.convergence_model.parametric()
        deflection_model = lens_system.lens_plane_model.convergence_model.deflection(
            parametric=True)
        forward_full = lens_system.get_forward_model_parametric()
        forward_source = lens_system.get_forward_model_parametric(
            only_source=True)
    else:
        convergence_model = lens_system.lens_plane_model.convergence_model
        deflection_model = lens_system.lens_plane_model.convergence_model.deflection()
        forward_full = lens_system.get_forward_model_full()
        forward_source = lens_system.get_forward_model_full(only_source=True)

    # Calculate the convergence and deflection field
    lens_mask = lens_system.lens_plane_model.space.get_mask_operator()
    lens_mask_slice = lens_system.lens_plane_model.space.get_mask(
        coordinates=True)
    convergence = lens_mask(convergence_model.model(position))
    convergence_mean = convergence_model.parametric()(position)
    perturbations = convergence_model.nonparametric()
    if perturbations is None or parametric:
        convergence_perturbations = np.ones_like(convergence)
    else:
        convergence_perturbations = perturbations(position)
    deflection = deflection_model.model(position).reshape(
        2, *lens_system.lens_plane_model.space.extend().shape
    )
    deflection_mean = deflection_mean.model(position).reshape(
        2, *lens_system.lens_plane_model.space.extend().shape
    )
    deflection = deflection[lens_mask_slice]
    deflection_mean = deflection_mean[lens_mask_slice]
    convergence_extent = lens_system.lens_plane_model.space.extend().extent

    # Calculate the model data
    # mask_lens = lens_system.lens_plane_model.space.get_mask()
    mask_lens_extent = lens_system.lens_plane_model.space.extent
    data_extent = {key: d.extent() for key, d in data.items()}
    data_model_full = R(forward_full(position))
    data_model_parametric = R(forward_parametric(position))
    source_light_lens = next(iter(forward_source(position).values()))
    source_light_lens = source_light_lens  # [mask_lens]
    if lens_light_model is None:
        first_row = data_model_parametric
        first_row_name = 'Parametric lensed light'
    else:
        first_row = lens_light_model(position)
        first_row_name = 'Lens light (no PSF)'

    fig, axes = plt.subplots(4, 5, figsize=(21, 14))
    row_imshow(
        axes[0],
        [source_mean, source_perturbations, source_reconstruction,
            input_source, input_source-source_reconstruction, ],
        ['Source parametric', 'Source perturbations', 'Source model',
         'Source input', 'Source (input-model)'],
        [*({'origin': 'lower', 'extent': source_extent,
            'vmin': 0, 'vmax': input_source.max()},)*4,
         {'origin': 'lower', 'extent': source_extent, 'cmap': 'RdBu_r',
          'vmin': -MINMAX, 'vmax': MINMAX}],
    )
    parameter_check(
        lens_system.source_plane_model.light_model, position, axes[0, 0])
    parameter_check(
        lens_system.source_plane_model.light_model, data[key].mock_pos,
        axes[0, 3])

    sdr = source_distortion_ratio(input_source, source_reconstruction)
    display_text(axes[0, -1], {
        's': f'SDR = {sdr:.2f}',
        'color': 'black',
        'bbox': {'facecolor': 'white', 'alpha': 0.5, 'edgecolor': 'none'},
    }, keyword='top_right')

    row_imshow(
        axes[1],
        [first_row,
         source_light_lens,
         data_model_full,
         data[key].data_2d,
         (data[key].data_2d-data_model_full)/data[key].noise_2d],
        [first_row_name,
         'Lensed source (no PSF)',
         'R(lens + lensed source)',
         'Data',
         '(data-reconstruction)/std'],
        [{'origin': 'lower', 'extent': data_extent[key]},
         {'origin': 'lower', 'extent': data_extent[key]},
         *({'origin': 'lower', 'extent': data_extent[key],
           'vmax': data[key].data_2d[~data[key].mask].max(),
            'vmin': data[key].data_2d[~data[key].mask].min()},)*2,
         {'origin': 'lower', 'extent': data_extent[key], 'cmap': 'RdBu_r',
         'vmin': -STANDARD_DEVIATIONS, 'vmax': STANDARD_DEVIATIONS}]
    )
    parameter_check(
        lens_system.lens_plane_model.light_model, position, axes[1, 0])

    conv_min = convergence_data[~np.isnan(convergence_data)].min()
    conv_max = convergence_data[~np.isnan(convergence_data)].max()
    if convergence_data.sum() == 0.:
        convergence_data = np.ones_like(convergence)
        conv_min = convergence.min()
        conv_max = convergence.max()
    row_imshow(
        axes[2],
        [convergence_mean,
         convergence_perturbations,
         convergence,
         convergence_data,
         convergence_data - convergence],
        ['Convergence parametric',
         'Convergence perturbations',
         'Convergence model',
         'Convergence data',
         'Convergence (data-model)', ],
        [*({'origin': 'lower',
            'cmap': 'RdYlBu_r',
            'extent': convergence_extent,
            'norm': LogNorm(conv_min, conv_max)},)*2,
         *({'origin': 'lower', 'cmap': 'RdYlBu_r', 'extent': mask_lens_extent,
            'norm': LogNorm(conv_min, conv_max)},)*2,
         {'origin': 'lower', 'cmap': 'RdBu_r', 'extent': mask_lens_extent,
          'vmin': -0.3, 'vmax': 0.3}]
    )
    parameter_check(
        lens_system.lens_plane_model.convergence_model, position, axes[2, 0])
    parameter_check(
        lens_system.lens_plane_model.convergence_model, data[key].mock_pos,
        axes[2, 3])

    defl_min = np.hypot(*deflection_data).min()
    defl_max = np.hypot(*deflection_data).max()
    if np.sum(deflection_data) == 0.:
        deflection_data = np.ones_like(deflection)
        defl_min = np.hypot(*deflection).min()
        defl_max = np.hypot(*deflection).max()

    row_imshow(
        axes[3],
        [np.hypot(*deflection_mean),
         np.hypot(*deflection_mean),
         np.hypot(*deflection),
         np.hypot(*deflection_data),
         np.hypot(*(deflection_data-deflection))],
        ['Deflection parametric',
         'Deflection parametric',
         'Deflection model',
         'Deflection data',
         'Deflection (data-model)'],
        [*({'origin': 'lower', 'extent': mask_lens_extent,
            'vmin': defl_min, 'vmax': defl_max},)*4,
         {'origin': 'lower', 'cmap': 'RdBu_r', 'extent': mask_lens_extent,
          'vmin': -0.3, 'vmax': 0.3}]
    )
    axes[3, 0].cla()

    compare = data[key].Ls if data[key].Ls is not None else data[key].data_2d
    hist_plot(axes[3, 0], compare, data_model_full)

    if output_name_template is None:
        plt.show()
        return None

    output_name = output_name_template.format(iteration=iteration)
    plt.tight_layout()
    plt.savefig(f'{output_name}.png')
    plt.close()


def standard_data_plot(
        position: Dict[str, ArrayLike],
        output_name_template: Optional[str],
        data: Dict[str, ImageData],
        lens_system: LensSystem,
        iteration: int,
        parametric: bool = True,
        log_data: bool = False,
) -> None:

    # Calculate the reconstructed source
    source_reconstruction = lens_system.source_plane_model.light_model(
        position).T  # FIXME: Why is here a transpose?
    source_extent = lens_system.source_plane_model.space.extend().extent

    convergence_model = lens_system.lens_plane_model.convergence_model
    deflection_model = lens_system.lens_plane_model.convergence_model.deflection(
        parametric=parametric)
    lens_light_model = lens_system.get_lens_light()
    if parametric:
        convergence_model = convergence_model.parametric()
        forward = lens_system.get_forward_model_parametric()
        forward_source = lens_system.get_forward_model_parametric(
            only_source=True)
    else:
        forward = lens_system.get_forward_model_full()
        forward_source = lens_system.get_forward_model_full(only_source=True)

    # Calculate the convergence and deflection field
    convergence = convergence_model.model(position)
    perturbations = convergence_model.nonparametric()
    if perturbations is None or parametric is True:
        convergence_perturbations = np.ones_like(convergence)
    else:
        convergence_perturbations = perturbations(position)
    deflection_field = deflection_model.model(position).reshape(
        2, *convergence.shape
    )
    convergence_extent = lens_system.lens_plane_model.space.extend().extent

    # Calculate the model data
    mask_lens_extent = lens_system.lens_plane_model.space.extent
    data_extent = {key: d.extent() for key, d in data.items()}

    sky = forward(position)
    sky.update(position)  # do this to get all parameters of the response model
    lensed_light = forward_source(position)
    lensed_light.update(position)

    if lens_light_model is None:
        lens_light = None
    else:
        lens_light = lens_light_model(position)

    fig, axes = plt.subplots(
        2 + len(data.keys()), 3, figsize=(19, 10 + 5 * len(data.keys()))
    )
    row_imshow(
        axes[0],
        [source_reconstruction, convergence, convergence_perturbations],
        ['Source reconstruction', 'Convergence', 'Convergence perturbations'],
        [{'origin': 'lower', 'extent': source_extent, 'vmin': 0},
         {'origin': 'lower', 'extent': convergence_extent, 'cmap': 'RdYlBu_r',
          'norm': LogNorm()},
         {'origin': 'lower', 'extent': convergence_extent, 'cmap': 'RdYlBu_r',
          'norm': LogNorm()}]
    )
    add_box(axes[0, 1], data_extent[next(iter(data.keys()))])
    add_box(axes[0, 2], data_extent[next(iter(data.keys()))])
    parameter_check(lens_system.source_plane_model.light_model,
                    position, axes[0, 0])
    parameter_check(lens_system.lens_plane_model.convergence_model,
                    position, axes[0, 1])

    if lens_light is None:
        row_imshow(
            axes[1],
            [source_reconstruction,
             lensed_light['sky'],
             np.hypot(*deflection_field)],
            ['Source reconstruction', 'Lensed Light', 'Deflection field'],
            [{'origin': 'lower',
              'extent': source_extent,
              'norm': LogNorm(vmin=np.max((1e-5, source_reconstruction.min())))
              },
             {'origin': 'lower', 'norm': LogNorm() if log_data else None,
              'extent': mask_lens_extent},
             {'origin': 'lower', 'extent': convergence_extent}])
        add_box(axes[1, 2], data_extent[next(iter(data.keys()))])
    else:
        row_imshow(
            axes[1],
            [source_reconstruction, lensed_light['sky'], lens_light],
            ['Source reconstruction', 'Lensed Light', 'Lens Light'],
            [{'origin': 'lower', 'extent': source_extent,
              'norm': LogNorm(vmin=source_reconstruction.max()*1e-3)},
             {'origin': 'lower',  # 'norm': LogNorm() if log_data else None,
              'extent': mask_lens_extent},
             {'origin': 'lower',  'norm': LogNorm() if log_data else None,
              'extent': mask_lens_extent}]
        )

    for ii, key in enumerate(data.keys()):
        data_max = data[key].data_2d[~data[key].mask].max()
        data_min = data[key].data_2d[~data[key].mask].min()
        if log_data:
            data_plot_kwargs = {'origin': 'lower', 'extent': data_extent[key],
                                'norm': LogNorm(vmin=1e-3, vmax=data_max)}
        else:
            data_plot_kwargs = {'origin': 'lower', 'extent': data_extent[key],
                                'vmax': data_max, 'vmin': data_min}

        data_model_noPsf = data[key].response_no_psf(sky)
        data_model_full = data[key].response(sky)

        plot_data = [
            data[key].data_2d,
            data_model_noPsf,
            (data[key].data_2d-data_model_full)/data[key].noise_2d]

        from .analysis_tools import chi2
        chi_2 = chi2(data[key].data_2d, data_model_full, data[key].noise_2d)

        plot_titles = ['Data', 'Model=Ls', f'abs(data-BLs)/std={chi_2:.2f}']
        plot_kwargs = [data_plot_kwargs, data_plot_kwargs,
                       {'origin': 'lower', 'extent': data_extent[key],
                        'cmap': 'RdBu_r', 'vmin': -STANDARD_DEVIATIONS,
                        'vmax': STANDARD_DEVIATIONS}]

        row_imshow(axes[ii+2], plot_data, plot_titles, plot_kwargs)
        if data[key].mask is not None:
            axes[ii+2, 0].contour(data[key].mask, levels=[0.5], colors='w',
                                  extent=data_extent[key])
            axes[ii+2, 2].contour(data[key].mask, levels=[0.5], colors='w',
                                  extent=data_extent[key])

    if output_name_template is None:
        plt.show()
    else:
        output_name = output_name_template.format(iteration=iteration)
        plt.tight_layout()
        plt.savefig(f'{output_name}.png')
        plt.close()


def samples_dictionary_parser(
    samples_config: Dict,
    lens_system: LensSystem,
) -> Dict[str, dict]:

    if samples_config is False:
        return {}

    samples_dict = {}
    for key, val in samples_config.items():
        if key in ['convergence']:
            samples_dict['_'.join(('samples', key))] = {
                'func': lens_system.lens_plane_model.convergence_model,
                'logscale': val['log_scale']}

        if key in ['source']:
            samples_dict['_'.join(('samples', key))] = {
                'func': lens_system.source_plane_model.light_model,
                'logscale': val['log_scale']}

    return samples_dict


def nifty_plotting_wrapper(plot_check: Callable) -> Callable:
    from functools import wraps

    @wraps(plot_check)
    def convert_to_jft_plot(samples: ift.SampleList, kl_iteration: int):
        from .nifty_connect import jft_samples_converter
        mean, _ = samples.sample_stat()
        jft_samples = jft_samples_converter(mean, samples)
        jft_state = jft.OptimizeVIState(nit=kl_iteration, key=42)
        return plot_check(jft_samples, jft_state)

    return convert_to_jft_plot


def plot_check_factory(
        outputdir: str,
        data: Dict,
        cfg: Dict,
        lens_system: LensSystem,
        parametric: bool,
        min_iterations: int = 0,
        operators_samples: Optional[dict] = {},
) -> Callable:

    plotter = partial(
        mock_data_plot if (
            cfg['mock'] or cfg['plotting_options'].get('mock_data', False)
        ) else standard_data_plot,

        output_name_template=f'{outputdir}/KL{{iteration}}' if outputdir is not None else None,
        data=data,
        lens_system=lens_system,
        parametric=parametric,
        log_data=cfg['plotting_options']['log_data']
    )

    def plot_check(samples: jft.Samples,
                   state: jft.OptimizeVIState):
        kl_iteration = state.nit + min_iterations
        print(f'Plotting iteration {kl_iteration} in {outputdir}\n')

        plotter(position=samples.pos.tree, iteration=kl_iteration)
        if operators_samples != {}:
            plot_sample_and_stats(
                output_directory=outputdir,
                operators_dict=operators_samples,
                res_sample_list=samples,
                state=state,
                iteration=kl_iteration,
                log_scale=cfg['plotting_options']['log_data'])

        print("Plotting done")

    return plot_check
